package edu.berkeley.qfat;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

// FIXME: recenter goal to have centroid coincident with tile
// FIXME: re-orient goal (how?)
//        fish case: ensure that spinal axis of fish is the x-axis of the tile

public class Main implements GLEventListener, MouseListener, MouseMotionListener, KeyListener, MouseWheelListener {
    
    boolean alt = false;
    boolean shift = false;
    boolean control = false;

    public void mouseWheelMoved(MouseWheelEvent e) {
        tz -= e.getWheelRotation();
    }

    public void keyTyped(KeyEvent e)  { }
    public void keyPressed(KeyEvent e)  {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = true; break;
            case KeyEvent.VK_ALT: alt = true; break;
            case KeyEvent.VK_SHIFT: shift = true; break;
        }
    }
    public void keyReleased(KeyEvent e) {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = false; break;
            case KeyEvent.VK_ALT: alt = false; break;
            case KeyEvent.VK_SHIFT: shift = false; break;
        }
    }

    public void mouseClicked(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mousePressed(MouseEvent e) { }
    public void mouseReleased(MouseEvent e) { }

    int mousex;
    int mousey;
    public void mouseMoved(MouseEvent e) {
        mousex = e.getX();
        mousey = e.getY();
    }

    float tx = 0;
    float ty = 0;
    float tz = 0;
    float anglex = 0;
    float angley = 0;
    public void mouseDragged(MouseEvent e) {
        if (shift) {
            tx += (mousex - e.getX())/(float)20;
            ty += (mousey - e.getY())/(float)20;
        } else {
            anglex -= mousex - e.getX();
            angley += mousey - e.getY();
        }
        mousex = e.getX();
        mousey = e.getY();
    }

    private Mesh tile = new Mesh();
    private Mesh goal = new Mesh();

    /** magnification factor */
    private static final float MAG = 1;

    Matrix[] translations;
    Mesh.Vert[] points;

    public Main(StlFile stlf) {

        for(int i=0; i<stlf.coordArray.length; i+=3) {
            Mesh.Vert p0 = goal.register(new Point(stlf.coordArray[i+0].x * MAG, stlf.coordArray[i+0].y * MAG, stlf.coordArray[i+0].z * MAG));
            Mesh.Vert p1 = goal.register(new Point(stlf.coordArray[i+1].x * MAG, stlf.coordArray[i+1].y * MAG, stlf.coordArray[i+1].z * MAG));
            Mesh.Vert p2 = goal.register(new Point(stlf.coordArray[i+2].x * MAG, stlf.coordArray[i+2].y * MAG, stlf.coordArray[i+2].z * MAG));
            Vec n  = new Vec(stlf.normArray[i/3].x * MAG, stlf.normArray[i/3].y  * MAG, stlf.normArray[i/3].z * MAG);
            Mesh.T t  = goal.newT(p0, p1, p2, n);
        }

        // rotate to align major axis -- this probably needs to be done by a human.
        goal.transform(new Matrix(new Vec(0, 0, 1), (float)(Math.PI/2)));


        float goal_width  = goal.diagonal().dot(new Vec(1, 0, 0));
        float goal_height = goal.diagonal().dot(new Vec(0, 1, 0));
        float goal_depth  = goal.diagonal().dot(new Vec(0, 0, 1));

        float width  = (float)0.6;
        float height = (float)0.08;
        float depth  = (float)0.3;
        translations = new Matrix[] {

            new Matrix(new Vec(-(width/2),  height,    0)),
            new Matrix(new Vec( (width/2),  height,    0)),
            new Matrix(new Vec(-(width/2), -height,    0)),
            new Matrix(new Vec( (width/2), -height,    0)),
            new Matrix(new Vec(-(width/2),       0,  depth)),
            new Matrix(new Vec( (width/2),       0,  depth)),
            new Matrix(new Vec(-(width/2),       0, -depth)),
            new Matrix(new Vec( (width/2),       0, -depth)),

            new Matrix(new Vec( width,           0,    0)),
            new Matrix(new Vec(-width,           0,    0)),
            /*
            new Matrix(new Vec(     0,           0,    depth)),
            new Matrix(new Vec(     0,           0,   -depth)),
            */
        };


        Mesh.Vert ltf = tile.register(new Point(-(width/2),  (height/2),  (depth/2)));
        Mesh.Vert mtf = tile.register(new Point( 0.0,        (height/2),  (depth/2)));
        Mesh.Vert rtf = tile.register(new Point( (width/2),  (height/2),  (depth/2)));
        Mesh.Vert ltn = tile.register(new Point(-(width/2),  (height/2), -(depth/2)));
        Mesh.Vert mtn = tile.register(new Point( 0.0,        (height/2), -(depth/2)));
        Mesh.Vert rtn = tile.register(new Point( (width/2),  (height/2), -(depth/2)));
        Mesh.Vert lbf = tile.register(new Point(-(width/2), -(height/2),  (depth/2)));
        Mesh.Vert mbf = tile.register(new Point( 0.0,       -(height/2),  (depth/2)));
        Mesh.Vert rbf = tile.register(new Point( (width/2), -(height/2),  (depth/2)));
        Mesh.Vert lbn = tile.register(new Point(-(width/2), -(height/2), -(depth/2)));
        Mesh.Vert mbn = tile.register(new Point( 0.0,       -(height/2), -(depth/2)));
        Mesh.Vert rbn = tile.register(new Point( (width/2), -(height/2), -(depth/2)));
        
        points = new Mesh.Vert[] {
            ltf,
            mtf,
            rtf,
            ltn,
            mtn,
            rtn,
            lbf,
            mbf,
            rbf,
            lbn,
            mbn,
            rbn
        };


        // top
        tile.newT(ltf, mtf, mtn, null);
        tile.newT(mtn, ltn, ltf, null);
        tile.newT(mtf, rtf, rtn, null);
        tile.newT(rtn, mtn, mtf, null);

        // bottom (swap normals)
        tile.newT(mbf, lbf, mbn, null);
        tile.newT(lbn, mbn, lbf, null);
        tile.newT(rbf, mbf, rbn, null);
        tile.newT(mbn, rbn, mbf, null);
        
        // left
        tile.newT(ltf, ltn, lbn, null);
        tile.newT(lbn, lbf, ltf, null);

        // right (swap normals)
        tile.newT(rtn, rtf, rbn, null);
        tile.newT(rbf, rbn, rtf, null);

        // front
        tile.newT(ltn, mtn, mbn, null);
        tile.newT(ltn, mbn, lbn, null);
        tile.newT(mtn, rtn, rbn, null);
        tile.newT(mtn, rbn, mbn, null);

        // back
        tile.newT(mtf, ltf, mbf, null);
        tile.newT(mbf, ltf, lbf, null);
        tile.newT(rtf, mtf, rbf, null);
        tile.newT(rbf, mtf, mbf, null);

        for(Matrix m : translations) {
            for(Mesh.T t1 : tile) {
                for(Mesh.T t2 : tile) {
                    if (t1==t2) continue;

                    if ((t1.v1().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e1().bind(t2.e3().pair);
                        t1.e2().bind(t2.e2().pair);
                        t1.e3().bind(t2.e1().pair);
                    }
                    if ((t1.v2().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v3().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e2().bind(t2.e3().pair);
                        t1.e3().bind(t2.e2().pair);
                        t1.e1().bind(t2.e1().pair);
                    }
                    if ((t1.v3().p.times(m).minus(t2.v1().p).mag() < Mesh.EPSILON) &&
                        (t1.v1().p.times(m).minus(t2.v3().p).mag() < Mesh.EPSILON) &&
                        (t1.v2().p.times(m).minus(t2.v2().p).mag() < Mesh.EPSILON)) {
                        t1.e3().bind(t2.e3().pair);
                        t1.e1().bind(t2.e2().pair);
                        t1.e2().bind(t2.e1().pair);
                    }
                }
            }
        }

        //xMesh.Vert mid = lbf.getE(mbn).shatter();

        // rescale to match volume
        float factor = (float)Math.pow(tile.volume() / goal.volume(), 1.0/3.0);
        goal.transform(new Matrix(factor));

        // translate to match centroid
        goal.transform(new Matrix(tile.centroid().minus(goal.centroid())));

        //tx.e2.shatter();
        //tx.e3.shatter();


        tile.bind();

        //mid.move(new Vec((float)0,0,(float)-0.05));
        //ltn.move(new Vec((float)0,0,(float)-0.05));

        //mtf.move(new Vec(0, (float)-0.05, (float)0.05));


        System.out.println("tile volume: " + tile.volume());
        System.out.println("goal volume: " + goal.volume());

        tile.score_against = goal;
        goal.score_against = tile;
    }
        Random random = new Random();

    public synchronized void breakit() {
        if (verts > 300) return;
        //double min = (tile.avgedge/tile.numedges)*(1+(4/(double)verts));
        //if (verts>0 && tile.es.peek().length() < min) return;
        PriorityQueue<Mesh.E> es = new PriorityQueue<Mesh.E>();
        for(Mesh.E e : tile.es) es.add(e);
        for(int i=0; i<10; i++) {
            Mesh.E e = es.poll();
            verts++;
            System.out.println("shatter " + e);
            e.shatter();
            tile.unbind();
            tile.bind();
        }
    }

    public synchronized void rand(double temperature, Mesh.Vert p) {
        double tile_score = tile.score();
        double goal_score = goal.score();
        
        p.rescore();
        //if (p.watch==null) return;
        float r1 = Math.abs(random.nextFloat());
        r1 = r1 - (float)Math.floor(r1);
        r1 = r1 * (float)0.01;
        r1 = r1 - (float)0.005;
        Vec v = p.watchback().p.minus(p.p).norm().times(r1);

        //v = p.norm().times(v.dot(p.norm()));

        boolean aspect = false;//(Math.abs(random.nextInt()) % 100) <= 2;
        Matrix old_tile_aspect = null;//goal.aspect;
        boolean good = true;
        if (aspect) {
            /*
            v = v.times(10);
            tile.aspect = new Matrix(tile.aspect.a / (v.x+1), tile.aspect.f / (v.y+1), tile.aspect.k / (v.z+1));
            tile.invaspect = new Matrix(1/tile.aspect.a, 1/tile.aspect.f, 1/tile.aspect.k);
            goal.rescore();
            tile.rescore();
            */
        } else {
            good = p.move(v);
        }
        double new_tile_score = tile.score();
        double new_goal_score = goal.score();
        double tile_delta = new_tile_score - tile_score;
        double goal_delta = new_goal_score - goal_score;
        double delta = tile_delta + goal_delta;
        double swapProbability = Math.exp((-1 * delta) / temperature);
        //boolean doSwap = Math.random() < swapProbability;
        boolean doSwap = good && (tile_delta <= 0 && goal_delta <= 0);
        if (doSwap) {
            tile_score = new_tile_score;
            goal_score = new_goal_score;
            //System.out.println("score: " + tile_score + " / " + goal_score);
            if (aspect) System.out.println("aspect " + v);
        } else {
            if (aspect) {
                //tile.aspect = old_tile_aspect;
                //tile.invaspect = new Matrix(1/tile.aspect.a, 1/tile.aspect.f, 1/tile.aspect.k);
                goal.rescore();
                tile.rescore();
            } else {
                p.move(v.times(-1));
            }
        }
    }

    /**
     * Take care of initialization here.
     */
    public void init(GLAutoDrawable gld) {
        GL gl = gld.getGL();
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        gl.glViewport(0, 0, 500, 300);
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glClearDepth(1.0);
        gl.glDepthFunc(GL.GL_LEQUAL);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        display(gld);
    }

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) { }
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { }
    public synchronized void display(GLAutoDrawable drawable) {
        GL gl = drawable.getGL();
        GLU glu = new GLU();
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glPointSize(5.0f);
        gl.glLoadIdentity();
        glu.gluPerspective(50-tz, ((float)drawable.getWidth())/drawable.getHeight(), 0.5, 10);
        glu.gluLookAt(0, 0, -1, 0, 0, 0, 0, 1, 0);
        gl.glTranslatef(tx/(float)20, ty/(float)20, 0);
        gl.glRotatef(anglex/3, 0, 1, 0);
        gl.glRotatef(angley/3, 1, 0, 0);

        gl.glBegin(GL.GL_TRIANGLES);
        draw(gl, true, tile);
        gl.glEnd();

        gl.glBegin(GL.GL_TRIANGLES);
        gl.glColor4f((float)0.5, (float)0.5, (float)0.5, (float)0.8);
        //draw(gl, false, goal);
        gl.glEnd();


        int i = 0;
        //gl.glDisable(GL.GL_DEPTH_TEST);
        gl.glColor4f(1,1,1,1);
        for(Matrix m : translations) {
            //if (v1.z==0 && v1.y==0) continue;
            i++;
            if (i != 1 /*&& i!=4*/) continue;
            Point p = new Point(0, 0, 0).times(m);
            Vec v = new Vec(p.x, p.y, p.z);
            v = v.times((float)1.04);
            gl.glTranslatef(v.x, v.y, v.z);
            draw(gl, false, tile);
            gl.glTranslatef(-v.x, -v.y, -v.z);
        }
        //gl.glEnable(GL.GL_DEPTH_TEST);
    }

    private synchronized void draw(GL gl, boolean triangles, Mesh mesh) {
        float red = 0.0f;
        float green = 0.0f;
        float blue = 0.0f;
        for(Mesh.T t : mesh) {
            if (red < 0.15) red = 1.0f;
            if (green < 0.15) green = 1.0f;
            if (blue < 0.15) blue = 1.0f;
            red -= .09f;
            green -= .12f;
            blue -= .15f;

            if (triangles) switch(t.color) {
                case 0: gl.glColor4f((float)0.25, (float)0.25, (float)0.75, (float)0.3); break;
                case 1: gl.glColor4f((float)0.25, (float)0.75, (float)0.25, (float)0.3); break;
                case 2: gl.glColor4f((float)0.75, (float)0.25, (float)0.25, (float)0.3); break;
                case 3: gl.glColor4f((float)0.50, (float)0.50, (float)0.50, (float)0.3); break;
            }
            //gl.glBegin(GL.GL_LINES);

            if (triangles) {
                gl.glBegin(GL.GL_TRIANGLES);
                t.glVertices(gl);
                gl.glEnd();
            } else {
                gl.glBegin(GL.GL_LINES);
                t.e1().p1.p.glVertex(gl);
                t.e1().p2.p.glVertex(gl);
                t.e2().p1.p.glVertex(gl);
                t.e2().p2.p.glVertex(gl);
                t.e3().p1.p.glVertex(gl);
                t.e3().p2.p.glVertex(gl);
                gl.glEnd();
            }

            Point centroid = t.centroid();
            gl.glBegin(GL.GL_LINES);
            gl.glColor3f(1, 1, 1);
            /*
            centroid.glVertex(gl);
            centroid.plus(t.norm().times(t.diameter())).glVertex(gl);
            */

            if (mesh==goal)
                for(Mesh.Vert p : new Mesh.Vert[] { t.v1(), t.v2(), t.v3() }) {
                    p.p.glVertex(gl);
                    //p.plus(p.norm().times(p.score()*10)).glVertex(gl);
                    p.partner().p.glVertex(gl);
                    //tile.nearest(p).centroid().glVertex(gl);
                }

            gl.glEnd();

        }
    }

    public static void main(String[] s) throws Exception {
        StlFile stlf = new StlFile();
        stlf.load("simplefish.stl");
        Main main = new Main(stlf);
        Frame f = new Frame();
        GLCapabilities glcaps = new GLCapabilities();
        GLCanvas glcanvas = new GLCanvas();
        glcanvas.addGLEventListener(main);
        f.add(glcanvas, BorderLayout.CENTER);
        f.pack();
        f.show();
        f.setSize(900, 900);
        f.doLayout();

        glcanvas.addMouseListener(main);
        glcanvas.addMouseMotionListener(main);
        glcanvas.addMouseWheelListener(main);
        glcanvas.addKeyListener(main);

        main.anneal(glcanvas);
    }
    public static int verts = 0;
    public void anneal(GLCanvas glcanvas) throws Exception {
        int verts = 0;
        while(true) {
            //Thread.sleep(10);
            for(int i=0; i<1; i++) {
                glcanvas.repaint();
                //tile.ts.get(Math.abs(random.nextInt()) % tile.ts.size()).e1().p1
                for(Mesh.T t : tile)
                    for(Mesh.Vert p : new Mesh.Vert[] { t.v1(), t.v2(), t.v3() }) {
                        rand(10,p);
                    }
                goal.rescore();
                tile.rescore();
            }
            breakit();
       }

    }

}