package edu.berkeley.qfat;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import edu.berkeley.qfat.geom.*;
import edu.wlu.cs.levy.CG.KDTree;
import edu.berkeley.qfat.geom.Point;

public class Mesh implements Iterable<Mesh.T> {

    private KDTree kd = new KDTree(3);

    public static float EPSILON = (float)0.0001;
    public static Random random = new Random();

    private HashMap<Point,Vert>  ps = new HashMap<Point,Vert>();
    public  HashSet<E>           es = new HashSet<E>();
    public  ArrayList<T>         ts = new ArrayList<T>();

    public Iterator<T> iterator() {
        if (es.size() == 0) return new FaceIterator();
        return new FaceIterator(es.iterator().next().p1);
    }

    public Point origin() { return new Point(0, 0, 0); }

    public Mesh score_against = null;
    public double score = 0;
    public float score() { return (float)score; }

    public int numedges = 0;
    public float avgedge = 0;

    public void unbind() {
        for(Mesh.T t : this) {
            t.v1().unbind();
            t.v2().unbind();
            t.v3().unbind();
        }
    }

    public void bind() {
        for(Mesh.T t : this) {
            t.e1().dobind();
            t.e2().dobind();
            t.e3().dobind();
        }
    }

    public float rescore() {
        int num = 0;
        double dist = 0;
        HashSet<Vert> done = new HashSet<Vert>();
        for(T t : this)
            for(Vert p : new Vert[] { t.v1(), t.v2(), t.v3() }) {
                if (done.contains(p)) continue;
                done.add(p);
                p.rescore();
            }
        for(T t : this)
            for(Vert p : new Vert[] { t.v1(), t.v2(), t.v3() })
                p.kdremove();
        kd = new KDTree(3);
        for(T t : this)
            for(Vert p : new Vert[] { t.v1(), t.v2(), t.v3() })
                p.kdinsert();
        return (float)(dist/num);
    }

    public void transform(Matrix m) {
        ArrayList<Vert> set = new ArrayList<Vert>();
        set.addAll(ps.values());
        for(Vert v : set) v.transform(m);
    }

    public Vec diagonal() {
        float min_x = Float.MAX_VALUE;
        float min_y = Float.MAX_VALUE;
        float min_z = Float.MAX_VALUE;
        float max_x = Float.MIN_VALUE;
        float max_y = Float.MIN_VALUE;
        float max_z = Float.MIN_VALUE;
        for(Point p : ps.keySet()) {
            if (p.x < min_x) min_x = p.x;
            if (p.y < min_y) min_y = p.y;
            if (p.z < min_z) min_z = p.z;
            if (p.x > max_x) max_x = p.x;
            if (p.y > max_y) max_y = p.y;
            if (p.z > max_z) max_z = p.z;
        }
        return new Vec(max_x - min_x, max_y - min_y, max_z - min_z);
    }

    public Point centroid() {
        float min_x = Float.MAX_VALUE;
        float min_y = Float.MAX_VALUE;
        float min_z = Float.MAX_VALUE;
        float max_x = Float.MIN_VALUE;
        float max_y = Float.MIN_VALUE;
        float max_z = Float.MIN_VALUE;
        for(Point p : ps.keySet()) {
            if (p.x < min_x) min_x = p.x;
            if (p.y < min_y) min_y = p.y;
            if (p.z < min_z) min_z = p.z;
            if (p.x > max_x) max_x = p.x;
            if (p.y > max_y) max_y = p.y;
            if (p.z > max_z) max_z = p.z;
        }
        return new Point((float)(max_x + min_x)/2,
                     (float)(max_y + min_y)/2,
                     (float)(max_z + min_z)/2);
    }

    public float volume() {
        double total = 0;
        for(T t : this) {
            double area = t.area();
            Vec origin_to_centroid = new Vec(new Point(0, 0, 0), t.centroid());
            boolean facingAway = t.norm().dot(origin_to_centroid) > 0;
            double height = Math.abs(t.norm().dot(origin_to_centroid));
            total += ((facingAway ? 1 : -1) * area * height) / 3.0;
        }
        return (float)total;
    }

    public Vert nearest(Point p) {
        Object[] results;
        try { results = kd.nearest(new double[]{p.x,p.y,p.z},1); } catch (Exception e) { throw new Error(e); }
        return (Vert)results[0];
    }

    public class BindingGroup {
        public HashSet<E> es = new HashSet<E>();
        public BindingGroup() { }
        public BindingGroup(E e) {
            es.add(e);
        }
        public void add(E e) {
            if (e.bg != null) { merge(e.bg); return; }
            es.add(e);
            e.bg = this;
        }
        public void merge(BindingGroup bg) {
            for(E e : bg.es) {
                e.bg = null;
                add(e);
            }
        }
    }

    public Vert register(Point p) { Vert v = ps.get(p); return v==null ? new Vert(p) : v; }
    public final class Vert {
        public Point p;
        private Vert(Point p) {
            this.p = p;
            if (ps.get(p) != null) throw new Error();
            ps.put(this.p, this);
        }
        public void kdremove() {
            if (!inserted) return;
            inserted = false;
            try { kd.delete(new double[]{p.x,p.y,p.z}); } catch (Exception e) { }
        }
        public void kdinsert() {
            if (inserted) return;
            inserted = true;
            try { kd.insert(new double[]{p.x,p.y,p.z},this); } catch (Exception e) { throw new Error(e); }
        }

        public float score() { return oldscore; }
        public void unscore() {
            if (watch == null) return;
            watch.watch_x -= p.x;
            watch.watch_y -= p.y;
            watch.watch_z -= p.z;
            watch.watch_count--;
            if (watch.watch_count==0) {
                watch.watch_x = 0;
                watch.watch_y = 0;
                watch.watch_z = 0;
            }
            watch = null;
        }
        public Vert partner() { return watch==null ? this : watch; }
        public Vert watchback() { return watch_count==0 ? partner() :
                register(new Point(watch_x/watch_count, watch_y/watch_count, watch_z/watch_count)); }
        public void rescore() {
            if (score_against == null) return;

            score -= oldscore;
            oldscore = 0;

            if (watch != null) unscore();
            Vert po = this;
            if (watch == null) {
                watch = score_against.nearest(po.p);

                // don't attract to vertices that face the other way
                if (watch.norm().dot(norm()) < 0) {
                    watch = null;
                } else {
                    watch.watch_x += po.p.x;
                    watch.watch_y += po.p.y;
                    watch.watch_z += po.p.z;
                    watch.watch_count++;
                }
            }

            double s1, s2;
            if (watch_count==0) s1 = 0;
            else                s1 = p.distance(watch_x/watch_count, watch_y/watch_count, watch_z/watch_count);
            s2 = watch==null ? 0 : po.p.distance(watch.p);
            oldscore = (float)(s1 + s2);
            score += oldscore;
        }

        /** does NOT update bound pairs! */
        public boolean transform(Matrix m) {
            // FIXME: screws up kdtree 
            // FIXME: screws up hashmap
            unscore();
            try {
                if (ps.get(this.p)==null) throw new Error();
                ps.remove(this.p);
                float newx = m.a*p.x + m.b*p.y + m.c*p.z + m.d;
                float newy = m.e*p.x + m.f*p.y + m.g*p.z + m.h;
                float newz = m.i*p.x + m.j*p.y + m.k*p.z + m.l;
                this.p = new Point(newx, newy, newz);
                // FIXME: what if we move onto exactly where another point is?
                ps.put(this.p,(Vert)this);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
            rescore();
            boolean good = true;
            /*
            for(T t : ts) {
                for(E e = this.e; ;) {
                    if (e.intersects(t)) { good = false; break; }
                    e = e.pair.next;
                    if (e == this.e) break;
                }
            }
            */
            /*
                if (t==this.t) continue;
                if (this.intersects(t)) good = false;
            }
            */
            return good;
        }
        public boolean move(Vec v) {
            Matrix m = new Matrix(v);
            Vert p = this;
            boolean good = true;
            do {
                good &= p.transform(m);
                v = v.times(binding); // bleh wrong
                p = p.bound_to;
            } while (p != this);
            return good;
        }

        public E makeE(Vert p2) {
            E e = getE(p2);
            if (e != null) return e;
            e = p2.getE(this);
            if (this.e == null && p2.e == null) return this.e = new E(this, p2);
            if (this.e == null && p2.e != null) return p2.makeE(this).pair;
            return new E(getFreeIncident(), p2);
        }

        public E getFreeIncident() {
            E ret = getFreeIncident(e, e);
            if (ret != null) return ret;
            ret = getFreeIncident(e.pair.next, e.pair.next);
            if (ret == null) throw new Error("unable to find free incident to " + this);
            return ret;
        }

        public E getFreeIncident(E start, E before) {
            E e = start;
            do {
                if (e.pair.p2 == this && e.pair.t == null && e.pair.next.t == null) return e.pair;
                e = e.pair.next;
            } while(e != before);
            return null;
        }

        public E getE(Vert p2) {
            E e = this.e;
            do {
                if (e==null) return null;
                if (e.p1 == this && e.p2 == p2) return e;
                e = e.pair.next;
            } while (e!=this.e);
            return null;
        }

        public boolean isBoundTo(Vert p) {
            Vert px = p;
            do {
                if (px==this) return true;
                px = px.bound_to;
            } while(px != p);
            return false;
        }

        public void unbind() { bound_to = this; binding = new Matrix(); }
        public void bind(Vert p) { bind(p, new Matrix()); }
        public void bind(Vert p, Matrix binding) {
            if (isBoundTo(p)) return;
            Vert temp_bound_to = p.bound_to;
            Matrix temp_binding = p.binding;
            p.bound_to = this.bound_to;
            p.binding = binding.times(this.binding); // FIXME: may have order wrong here
            this.bound_to = temp_bound_to;
            this.binding = temp_binding.times(temp_binding); // FIXME: may have order wrong here
        }
        public Vec norm() {
            Vec norm = new Vec(0, 0, 0);
            E e = this.e;
            do {
                if (e.t != null) norm = norm.plus(e.t.norm().times((float)e.prev.angle()));
                e = e.pair.next;
            } while(e != this.e);
            return norm.norm();
        }

        Vert bound_to = this;
        int watch_count;
        float watch_x;
        float watch_y;
        float watch_z;
        Vert watch;
        E e;                // some edge *leaving* this point
        Matrix binding = new Matrix();
        float oldscore = 0;
        boolean inserted = false;
    }

    /** [UNIQUE] an edge */
    public final class E implements Comparable<E> {

        public boolean intersects(T t) {
            double A0=t.v1().p.x, A1=t.v1().p.y, A2=t.v1().p.z;
            double B0=t.v2().p.x, B1=t.v2().p.y, B2=t.v2().p.z;
            double C0=t.v3().p.x, C1=t.v3().p.y, C2=t.v3().p.z;
            double j0=p1.p.x, j1=p1.p.y, j2=p1.p.z;
            double k0=p2.p.x, k1=p2.p.y, k2=p2.p.z;
            double J0, J1, J2;
            double K0, K1, K2;
            double i0, i1, i2;
            double a0, a1, a2;
            double b0, b1, b2;
            double c0, c1, c2;
            double in_det;
            double R00, R01, R02, R03,
                R10, R11, R12, R13,
                R20, R21, R22, R23,
                R30, R31, R32, R33;


            /* a = B - A */
            a0 = B0 - A0; 
            a1 = B1 - A1; 
            a2 = B2 - A2;
            /* b = C - B */
            b0 = C0 - A0;
            b1 = C1 - A1;
            b2 = C2 - A2;
            /* c = a &times; b */
            c0 = a1 * b2 - a2 * b1;
            c1 = a2 * b0 - a0 * b2;
            c2 = a0 * b1 - a1 * b0;
 
            /* M^(-1) = (1/det(M)) * adj(M) */
            in_det = 1 / (c0 * c0 + c1 * c1 + c2 * c2);
            R00 = (b1 * c2 - b2 * c1) * in_det;
            R01 = (b2 * c0 - b0 * c2) * in_det;
            R02 = (b0 * c1 - b1 * c0) * in_det;
            R10 = (c1 * a2 - c2 * a1) * in_det;
            R11 = (c2 * a0 - c0 * a2) * in_det;
            R12 = (c0 * a1 - c1 * a0) * in_det;
            R20 = (c0) * in_det;
            R21 = (c1) * in_det;
            R22 = (c2) * in_det;
  
            /* O = M^(-1) * A */
            R03 = -(R00 * A0 + R01 * A1 + R02 * A2);
            R13 = -(R10 * A0 + R11 * A1 + R12 * A2);
            R23 = -(R20 * A0 + R21 * A1 + R22 * A2);
 
            /* fill in last row of 4x4 matrix */
            R30 = R31 = R32 = 0;
            R33 = 1;
  
            J2 = R20 * j0 + R21 * j1 + R22 * j2 + R23;
            K2 = R20 * k0 + R21 * k1 + R22 * k2 + R23;
            if (J2 * K2 >= 0) return false;

            J0 = R00 * j0 + R01 * j1 + R02 * j2 + R03;
            K0 = R00 * k0 + R01 * k1 + R02 * k2 + R03;
            i0 = J0 + J2 * ((K0 - J0) / (J2 - K2));
            if (i0 < 0 || i0 > 1) return false;
  
            J1 = R10 * j0 + R11 * j1 + R12 * j2 + R13;
            K1 = R10 * k0 + R11 * k1 + R12 * k2 + R13;
            i1 = J1 + J2 * ((K1 - J1) / (J2 - K2));
            if (i1 < 0 || i1 > 1 || i0 + i1 > 1) return false;

            return true;            
        }

        public int compareTo(E e) {
            return e.length() > length() ? 1 : -1;
        }

        public final Vert p1, p2;
        T t;     // triangle to our "left"
        E prev;  // previous half-edge
        E next;  // next half-edge
        E pair;  // partner half-edge


        public BindingGroup bg = new BindingGroup(this);

        public void bind(E e) { bind(e, new Matrix()); }
        public void bind(E e, Matrix m) { e.bg.add(this); }

        public void dobind() {
            if (bg==null) return;
            for(E ex : bg.es) {
                if (ex==this) continue;
                p1.bind(ex.p1);
                p2.bind(ex.p2);
            }
        }

        boolean shattered = false;
        public Vert shatter() { return shatter(register(midpoint()), null, null); }
        public Vert shatter(Vert mid, BindingGroup bg1, BindingGroup bg2) {
            if (shattered) return mid;
            shattered = true;

            Vert r = next.p2;
            E next = this.next;
            E prev = this.prev;

            if (bg1==null) bg1 = new BindingGroup();
            if (bg2==null) bg2 = new BindingGroup();
            for(E e : bg.es) e.shatter(register(e.midpoint()), bg1, bg2);
            pair.shatter();
            destroy();

            newT(r, p1, mid, null);
            newT(r, mid, p2, null);
            bg1.add(p1.getE(mid));
            bg2.add(mid.getE(p2));
            return mid;
        }

        public boolean destroyed = false;
        public void destroy() {
            if (destroyed) return;
            destroyed = true;
            pair.destroyed = true;
            if (next.t != null) next.t.destroy();
            if (prev.t != null) prev.t.destroy();
            if (pair.next.t != null) ts.remove(pair.next.t);
            if (pair.prev.t != null) ts.remove(pair.prev.t);
            next.t = null;
            prev.t = null;
            pair.next.t = null;
            pair.prev.t = null;
            this.bg = null;
            pair.bg = null;
            pair.prev.next = next;
            next.prev = pair.prev;
            prev.next = pair.next;
            pair.next = prev;
            if (p1.e == this) p1.e = prev.next;
            if (pair.p1.e == pair) pair.p1.e = pair.prev.next;
            es.remove(this);
            es.remove(pair);
            avgedge -= this.length();
            avgedge -= pair.length();
            numedges--;
            numedges--;
        }

        private void sync() {
            this.prev.next = this;
            this.next.prev = this;
            this.pair.pair = this;
            if (this.next.p1 != p2) throw new Error();
            if (this.prev.p2 != p1) throw new Error();
            if (this.p1.e == null) this.p1.e = this;
            es.add(this);
            if (!added) {
                added = true;
                numedges++;
                avgedge += length();
            }
        }
        private boolean added = false;

        public T makeT() { return t==null ? (t = new T(this)) : t; }

        /** angle between this half-edge and the next */
        public double angle() {
            Vec v1 = next.p2.p.minus(p2.p);
            Vec v2 = this.p1.p.minus(p2.p);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        public void makeAdjacent(E e) {
            if (this.next == e) return;
            if (p2 != e.p1) throw new Error("cannot make adjacent -- no shared vertex");
            if (t != null || e.t != null) throw new Error("cannot make adjacent -- edges not both free");

            E freeIncident = p2.getFreeIncident(e, this);

            e.prev.next = freeIncident.next;
            freeIncident.next.prev = e.prev;

            freeIncident.next = this.next;
            this.next.prev = freeIncident;
            
            this.next = e;
            e.prev = this;

            sync();
            freeIncident.sync();
        }

        /** creates an isolated edge out in the middle of space */
        public E(Vert p1, Vert p2) {
            if (p1==p2) throw new Error("attempt to create edge with single vertex: " + p1);
            this.p1 = p1;
            this.p2 = p2;
            this.prev = this.next = this.pair = new E(this, this, this);
            sync();
        }

        /** adds a new half-edge from prev.p2 to p2 */
        public E(E prev, Vert p2) {
            this.p1 = prev.p2;
            this.p2 = p2;
            this.prev = prev;
            if (p2.getE(p1) != null) throw new Error();
            if (p2.e==null) {
                this.next = this.pair = new E(this, this, prev.next);
            } else {
                E q = p2.getFreeIncident();
                this.next = q.next;
                this.next.prev = this;
                E z = prev.next;
                this.prev.next = this;
                this.pair = new E(q, this, z);
            }
            sync();
        }

        /** adds a new half-edge to the mesh with a given predecessor, successor, and pair */
        public E(E prev, E pair, E next) {
            this.p1 = prev.p2;
            this.p2 = next.p1;
            this.prev = prev;
            this.next = next;
            this.pair = pair;
            sync();
        }
        public Point midpoint() { return new Point((p1.p.x+p2.p.x)/2, (p1.p.y+p2.p.y)/2, (p1.p.z+p2.p.z)/2); }
        public boolean has(Vert v) { return v==p1 || v==p2; }
        public float length() { return p1.p.minus(p2.p).mag(); }
        public String toString() { return p1+"->"+p2; }
    }

    public T newT(Vert p1, Vert p2, Vert p3, Vec norm) {
        if (norm != null) {
            Vec norm2 = p3.p.minus(p1.p).cross(p2.p.minus(p1.p));
            float dot = norm.dot(norm2);
            //if (Math.abs(dot) < EPointSILON) throw new Error("dot products within epsilon of each other: "+norm+" "+norm2);
            if (dot < 0) { Vert p = p1; p1=p2; p2 = p; }
        }
        E e12 = p1.makeE(p2);
        E e23 = p2.makeE(p3);
        E e31 = p3.makeE(p1);
        while(e12.next != e23 || e23.next != e31 || e31.next != e12) {
            e12.makeAdjacent(e23);
            e23.makeAdjacent(e31);
            e31.makeAdjacent(e12);
        }
        T ret = e12.makeT();
        if (e12.t == null) throw new Error();
        if (e23.t == null) throw new Error();
        if (e31.t == null) throw new Error();
        return ret;
    }


    public class FaceIterator implements Iterator<T> {
        private HashSet<T> visited = new HashSet<T>();
        private LinkedList<T> next = new LinkedList<T>();
        public FaceIterator() { }
        public FaceIterator(Vert v) { next.addFirst(v.e.t); }
        public boolean hasNext() { return next.peek()!=null; }
        public void remove() { throw new Error(); }
        public T next() {
            T ret = next.removeFirst();
            if (ret == null) return null;
            visited.add(ret);
            T t1 = ret.e1().pair.t;
            T t2 = ret.e2().pair.t;
            T t3 = ret.e3().pair.t;
            if (t1 != null && !visited.contains(t1)) next.addFirst(t1);
            if (t2 != null && !visited.contains(t2)) next.addFirst(t2);
            if (t3 != null && !visited.contains(t3)) next.addFirst(t3);
            return ret;
        }
    }

    /** [UNIQUE] a triangle (face) */
    public final class T extends Triangle {
        public final E e1;
        public final int color;

        public void destroy() {
            ts.remove(this);
        }

        T(E e1) {
            this.e1 = e1;
            E e2 = e1.next;
            E e3 = e2.next;
            if (e1==e2 || e1==e3) throw new Error();
            if (e3.next!=e1) throw new Error();
            if (e1.t!=null || e2.t!=null || e3.t!=null) throw new Error("non-manifold surface or disagreeing normals");
            e1.t = this;
            e1.next.t = this;
            e1.next.next.t = this;

            // FIXME: check for sealed/watertight surface once construction is complete (and infer normal(s)?)

            int color = Math.abs(random.nextInt());
            while(true) {
                color = color % 4;
                if (e1().pair.t != null && color == e1().pair.t.color) { color++; continue; }
                if (e2().pair.t != null && color == e2().pair.t.color) { color++; continue; }
                if (e3().pair.t != null && color == e3().pair.t.color) { color++; continue; }
                break;
            }
            this.color = color;

            // FIXME unnecssary?
            ts.add(this);
            v1().kdinsert();
            v2().kdinsert();
            v3().kdinsert();
        }
        public E e1() { return e1; }
        public E e2() { return e1.next; }
        public E e3() { return e1.prev; }
        public Vert v1() { return e1.p1; }
        public Vert v2() { return e1.p2; }
        public Vert v3() { return e1.next.p2; }
        public Point p1() { return e1.p1.p; }
        public Point p2() { return e1.p2.p; }
        public Point p3() { return e1.next.p2.p; }
        public boolean hasE(E e) { return e1==e || e1.next==e || e1.prev==e; }
        public boolean has(Vert v) { return v1()==v || v2()==v || v3()==v; }
    }

}
